# flexiai/core/flexi_managers/stream_event_handler.py
import json
from openai import AssistantEventHandler



class StreamEventHandler(AssistantEventHandler):
    def __init__(self, client, logger, run_manager, assistant_name):
        """
        Initializes the StreamEventHandler with the necessary dependencies.
        """
        self.client = client
        self.logger = logger
        self.run_manager = run_manager
        self.assistant_name = assistant_name  # Accept assistant_name here
        self.printed_message_ids = set()      # Track printed message IDs
        super().__init__()  # Call the parent class constructor


    def set_assistant_name(self, name):
        """
        Update the assistant's name dynamically.
        """
        self.assistant_name = name


    def _init(self, stream):
        """
        This method initializes the stream for the event handler.
        """
        self.__stream = stream
        super()._init(stream)


    def on_event(self, event):
        """
        Handles events coming from the streaming API.
        Specifically, it triggers when an event with 'requires_action' or message content is received.
        """
        if hasattr(event, 'event') and event.event == "thread.run.requires_action":
            run_id = event.data.id  # Retrieve the run ID from the event data
            self.logger.info(f"Run requires action: {run_id}")
            self.stream_handle_requires_action(event.data, run_id)
        elif hasattr(event, 'event') and event.event == "thread.message.delta":
            # Pass the event to the message handler and print the assistant's response
            self.handle_message_delta(event)


    def handle_message_delta(self, event):
        """
        Handles incoming message deltas (streaming content) and prints the assistant's responses immediately.
        """
        # Log the entire event to see what data is being passed in
        self.logger.info(f"Received event: {event}")

        # Print assistant name only once at the start of a new message
        if event.data.id not in self.printed_message_ids:
            self.logger.info(f"Processing message with ID: {event.data.id}")  # Debug log for message ID
            print(f"\n{self.assistant_name}: ", end='', flush=True)
            self.printed_message_ids.add(event.data.id)  # Mark this message as printed with assistant name

        # Process the message delta and print content immediately
        if hasattr(event.data.delta, 'content'):
            for block in event.data.delta.content:
                if block.type == 'text' and hasattr(block.text, 'value'):
                    # Print each chunk directly as it arrives
                    chunk = block.text.value
                    self.logger.info(f"Received message chunk: {chunk}")  # Debug log for message chunk
                    print(chunk, end='', flush=True)
                else:
                    self.logger.info("No valid content found in this message block.")
        else:
            self.logger.info("No delta content found in the event.")

        # Once the message is completed, ensure newline after the response
        if hasattr(event, 'event') and event.event == "thread.message.completed":
            self.logger.info(f"Message completed: {event.data.id}")  # Log message completion
            print()  # Ensure new line after the assistant's response completes


    def start_streaming_response(self, thread_id, assistant_id):
        """
        Start streaming response from the assistant and handle the streaming process.
        Uses the instance of StreamEventHandler.
        """
        self.logger.info(f"Starting assistant run for thread {thread_id} with assistant {assistant_id}.")

        # Start the stream with the event handler
        try:
            with self.client.beta.threads.runs.stream(
                thread_id=thread_id,
                assistant_id=assistant_id,
                event_handler=self  # Use the current instance (self) as the event handler
            ) as stream:
                self.logger.info("Stream started successfully.")  # Log stream start
                # Process each event as it arrives in the stream
                for event in stream:
                    self.on_event(event)  # Handle each event using the on_event method
        except Exception as e:
            self.logger.error(f"Unexpected error during streaming: {str(e)}", exc_info=True)


    def stream_handle_requires_action(self, data, run_id):
        """
        Handles the required actions for a given run using event streaming.
        Submits tool outputs cleanly without logging in the streamed content.
        """
        try:
            tool_calls = data.required_action.submit_tool_outputs.tool_calls
            tool_outputs = []

            # Check if we need to execute tool calls in parallel or sequentially
            use_parallel = len(tool_calls) > 1

            if use_parallel:
                self.logger.info("Executing tool calls in parallel mode.")
                try:
                    tasks = [
                        {
                            'function_name': tool_call.function.name,
                            'parameters': json.loads(tool_call.function.arguments)
                        }
                        for tool_call in tool_calls
                    ]
                    results = self.run_manager.call_parallel_functions(tasks)

                    for tool_call, result in zip(tool_calls, results):
                        tool_outputs.append(self.run_manager.prepare_tool_output(tool_call, result))
                except Exception as e:
                    self.logger.error(f"Error during parallel execution: {str(e)}", exc_info=True)
                    raise
            else:
                # Sequential execution of tool calls
                for tool_call in tool_calls:
                    try:
                        function_name = tool_call.function.name
                        arguments = json.loads(tool_call.function.arguments)

                        action_type = self.run_manager.determine_action_type(function_name)

                        if action_type == "call_assistant":
                            result = self.run_manager.call_assistant_with_arguments(function_name, **arguments)
                        else:
                            result = self.run_manager.execute_personal_function_with_arguments(function_name, **arguments)

                        tool_outputs.append(self.run_manager.prepare_tool_output(tool_call, result))
                    except Exception as e:
                        tool_outputs.append(self.run_manager.prepare_tool_output(tool_call, e, success=False))
                        self.logger.error(f"Error executing tool call {tool_call.id}: {str(e)}", exc_info=True)

            # After tool outputs are generated, stream them properly without mixing logs
            self.stream_submit_tool_outputs(tool_outputs, run_id)

        except Exception as e:
            self.logger.error(f"Error handling action for run {run_id}: {str(e)}", exc_info=True)


    def stream_submit_tool_outputs(self, tool_outputs, run_id):
        """
        Submits the tool outputs using a streaming mechanism. Ensure that the streaming
        prints only the assistant response and not logging information.
        """
        try:
            with self.client.beta.threads.runs.submit_tool_outputs_stream(
                thread_id=self.current_run.thread_id,
                run_id=self.current_run.id,
                tool_outputs=tool_outputs,
                event_handler=self  # Reuse the current event handler
            ) as stream:
                for text_delta in stream.text_deltas:
                    if text_delta.content:
                        print(text_delta.content, end='', flush=True)
        except Exception as e:
            self.logger.error(f"Error submitting tool outputs for run {run_id}: {str(e)}", exc_info=True)
